#include <stdint.h>

#include "log.h"

#include "nmea.h"
#include "gps.h"
#include "mstimer.h"
#include "clk.h"
#include "clkutc.h"
#include "clktime.h"
#include "pps.h"

#define TIMEOUT_MS    10000
#define CONFIDENCE_OK    30

static time64   lastSeconds = 0;
static uint32_t lastMs      = 0;
uint32_t NmeaTimeGetMs()      { return lastMs; }
time64   NmeaTimeGetSeconds() { return lastSeconds; }

//Stability confidence
static int confidence =  0;
static void confidenceReset()
{
    if (confidence == CONFIDENCE_OK) GpsLog("NMEA time is not stable\r\n");
    confidence = 0;
}
static void confidenceIncrease()
{
    if (confidence == CONFIDENCE_OK) return;
    confidence++;
    if (confidence == CONFIDENCE_OK) GpsLog("NMEA time is stable\r\n");
}

bool NmeaTimeIsStable() { return confidence == CONFIDENCE_OK; }

static uint32_t msTimerMsgTimeout = (uint32_t)-TIMEOUT_MS;
static bool nmeaWasStopped = true;

void NmeaTimeReceived(time64 nmeaMsgTime, int nmeaMs)
{   
    //Reset the msg timeout timer
    static bool lastSet = false;
    if (nmeaWasStopped)
    {
        if (lastSet) GpsLog("NMEA time messages have restarted after %d ms\r\n", MsTimerCount - msTimerMsgTimeout);
        else         GpsLog("NMEA time messages have started\r\n");
    }
    nmeaWasStopped = false;
    msTimerMsgTimeout = MsTimerCount;

    //Update confidence    
    if (lastSet)
    {
        time64 interval = nmeaMsgTime - lastSeconds;
        if (interval == 1)
        {
            confidenceIncrease();
        }
        else
        {
            GpsLog("NMEA time message has jumped %lld seconds\r\n", interval);
            confidenceReset();
        }
    }
    
    //Save the timings
   lastSeconds = nmeaMsgTime;
   lastMs      = MsTimerCount;
   lastSet     = true;
}
void NmeaTimeMain()
{
    //Check if messages have stopped
    if (MsTimerRelative(msTimerMsgTimeout, TIMEOUT_MS))
    {
        if (!nmeaWasStopped) GpsLog("NMEA time messages have stopped\r\n");
        nmeaWasStopped = true;
        confidenceReset();
    }    
    
    //Disable syncing until Nmea module is ready
    if (NmeaModuleReadiness != NMEA_READINESS_READY)
    {
        confidenceReset();
    }
}
void NmeaTimeInit()
{
    GpsLog("NMEA time reset\r\n");
    confidenceReset();
}
